/***************************************************************************
 *cr
 *cr            (C) Copyright 2007 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <inttypes.h>
#include "parboil.h"
#include <cuda.h>

#include "sad.h"
#include "largerBlocks.h"
#include "file.h"
#include "image.h"
 
#include "sad4.cu"
#include "largerBlocks.cu"

#define CUDA_ERRCK \
  {cudaError_t err = cudaGetLastError(); \
    if (err) fprintf(stderr, "CUDA error: %s\n", cudaGetErrorString(err)); \
  }

static void
write_sads(char *filename,
	   int image_width_macroblocks,
	   int image_height_macroblocks,
	   unsigned short *sads);

/* FILE I/O */


/* Compare the reference SADs to the expected SADs.
 */
void
check_sads(unsigned short *sads_reference,
	   unsigned short *sads_computed,
	   int image_size_macroblocks)
{
  int block;

  /* Check the 4x4 SAD values.  These are in sads_reference.
   * Ignore the data at the beginning of sads_computed. */
  sads_computed += 25 * MAX_POS_PADDED * image_size_macroblocks;

  for (block = 0; block < image_size_macroblocks; block++)
    {
      int subblock;

      for (subblock = 0; subblock < 16; subblock++)
	{
	  int sad_index;

	  for (sad_index = 0; sad_index < MAX_POS; sad_index++)
	    {
	      int index =
		(block * 16 + subblock) * MAX_POS_PADDED + sad_index;

	      if (sads_reference[index] != sads_computed[index])
		{
#if 0
		  /* Print exactly where the mismatch was seen */
		  printf("M %3d %2d %4d (%d = %d)\n", block, subblock, sad_index, sads_reference[index], sads_computed[index]);
#else
		  goto mismatch;
#endif
		}
	    }
	}
    }

  printf("Success.\n");
  return;

 mismatch:
  printf("Computed SADs do not match expected values.\n");
}

/* Extract the SAD data for a particular block type for a particular
 * macroblock from the array of SADs of that block type. */
static inline void
write_subblocks(FILE *outfile, unsigned short *subblock_array, int macroblock,
		int count)
{
  int block;
  int pos;

  for (block = 0; block < count; block++)
    {
      unsigned short *vec = subblock_array +
	(block + macroblock * count) * MAX_POS_PADDED;

      /* Write all SADs for this sub-block */
      for (pos = 0; pos < MAX_POS; pos++)
	write16u(outfile, *vec++);
    }
}

/* Write some SAD data to a file for output checking.
 *
 * All SAD values for six rows of macroblocks are written.
 * The six rows consist of the top two, middle two, and bottom two image rows.
 */
void
write_sads(char *filename,
	   int mb_width,
	   int mb_height,
	   unsigned short *sads)
{
  FILE *outfile = fopen(filename, "w");
  int mbs = mb_width * mb_height;
  int row_indir;
  int row_indices[6] = {0, 1,
			mb_height / 2 - 1, mb_height / 2,
			mb_height - 2, mb_height - 1};

  if (outfile == NULL)
    {
      fprintf(stderr, "Cannot open output file\n");
      exit(-1);
    }

  /* Write the number of output macroblocks */
  write32u(outfile, mb_width * 6);

  /* Write zeros */
  write32u(outfile, 0);

  /* Each row */
  for (row_indir = 0; row_indir < 6; row_indir++)
    {
      int row = row_indices[row_indir];

      /* Each block in row */
      int block;
      for (block = mb_width * row; block < mb_width * (row + 1); block++)
	{
	  int blocktype;

	  /* Write SADs for all sub-block types */
	  for (blocktype = 1; blocktype <= 7; blocktype++)
	    write_subblocks(outfile,
			    sads + SAD_TYPE_IX(blocktype, mbs),
			    block,
			    SAD_TYPE_CT(blocktype));
	}
    }

  fclose(outfile);
}

/* FILE I/O for debugging */



/* MAIN */

int
main(int argc, char **argv)
{
  struct image_i16 *ref_image;
  struct image_i16 *cur_image;
  unsigned short *sads_computed; /* SADs generated by the program */

  int image_size_bytes;
  int image_width_macroblocks, image_height_macroblocks;
  int image_size_macroblocks;

  //struct pb_TimerSet timers;
  //struct pb_Parameters *params;
  
  /* Initialize the parameters structure */
  struct pb_Parameters *params = (struct pb_Parameters *)malloc(sizeof(struct pb_Parameters));
  params->outFile = NULL;
  params->inpFiles = (char **)malloc(sizeof(char *)*2);
  params->inpFiles[0] = NULL;
  
// Read input from command line
#ifdef SIZE0
  params->inpFiles[0] = "~/software/parboil-2.5/datasets/sad/default/input/reference.bin";
  params->inpFiles[1] = "~/software/parboil-2.5/datasets/sad/default/input/frame.bin";
#endif
#ifdef SIZE1
  params->inpFiles[0] = "~/software/parboil-2.5/datasets/sad/large/input/reference.bin";
  params->inpFiles[1] = "~/software/parboil-2.5/datasets/sad/large/input/frame.bin";
#endif
  //pb_InitializeTimerSet(&timers);
  //params = pb_ReadParameters(&argc, argv);
/*
  if (pb_Parameters_CountInputs(params) != 2)
    {
      fprintf(stderr, "Expecting two input filenames\n");
      exit(-1);
    }
*/
  /* Read input files */
  //pb_SwitchToTimer(&timers, pb_TimerID_IO);
  ref_image = load_image(params->inpFiles[0]);
  cur_image = load_image(params->inpFiles[1]);
  //pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);

  if ((ref_image->width != cur_image->width) ||
      (ref_image->height != cur_image->height))
    {
      fprintf(stderr, "Input images must be the same size\n");
      exit(-1);
    }
  if ((ref_image->width % 16) || (ref_image->height % 16))
    {
      fprintf(stderr, "Input image size must be an integral multiple of 16\n");
      exit(-1);
    }

  /* Compute parameters, allocate memory */
  image_size_bytes = ref_image->width * ref_image->height * sizeof(short);
  image_width_macroblocks = ref_image->width >> 4;
  image_height_macroblocks = ref_image->height >> 4;
  image_size_macroblocks = image_width_macroblocks * image_height_macroblocks;
  
  sads_computed = (unsigned short *)
    malloc(41 * MAX_POS_PADDED * image_size_macroblocks * sizeof(short));

  /* Run the kernel code */
  {
    struct cudaArray *ref_ary;  /* Reference image on the device */
    short *d_cur_image;         /* Current image on the device */
    unsigned short *d_sads;     /* SADs on the device */
    dim3 macroblock_grid(image_width_macroblocks, image_height_macroblocks);

    //pb_SwitchToTimer(&timers, pb_TimerID_COPY);
    cudaMalloc((void **)&d_cur_image, image_size_bytes);
    CUDA_ERRCK
    cudaMallocArray(&ref_ary, &get_ref().channelDesc,
                    ref_image->width, ref_image->height);
    CUDA_ERRCK

    /* Transfer current image to device */
    cudaMemcpy(d_cur_image, cur_image->data, image_size_bytes,
               cudaMemcpyHostToDevice);
    CUDA_ERRCK

    /* Transfer reference image to device */
    cudaMemcpy2DToArray(ref_ary,
                        0, 0,
                        ref_image->data,
                        ref_image->width * sizeof(unsigned short),
                        ref_image->width * sizeof(unsigned short),
                        ref_image->height,
                        cudaMemcpyHostToDevice);
    CUDA_ERRCK
    cudaBindTextureToArray(get_ref(), ref_ary);
    CUDA_ERRCK

    /* Allocate SAD data on the device */
    cudaMalloc((void **)&d_sads, 41 * MAX_POS_PADDED * image_size_macroblocks *
	       sizeof(unsigned short));
    CUDA_ERRCK
    cudaMemset(d_sads, 0, 41 * MAX_POS_PADDED * image_size_macroblocks *
	       sizeof(unsigned short));
    CUDA_ERRCK

    //pb_SwitchToTimer(&timers, pb_TimerID_KERNEL);

    /* Run the 4x4 kernel */
    mb_sad_calc<<<dim3(CEIL(ref_image->width / 4, THREADS_W),       CEIL(ref_image->height / 4, THREADS_H)),                    dim3(CEIL(MAX_POS, POS_PER_THREAD) * THREADS_W * THREADS_H),                           SAD_LOC_SIZE_BYTES>>>
      (d_sads,
       (unsigned short *)d_cur_image,
       image_width_macroblocks,
       image_height_macroblocks);
    CUDA_ERRCK

    // Run the larger-blocks kernels
    larger_sad_calc_8<<<macroblock_grid, dim3(32, 4)>>>
      (d_sads,
       image_width_macroblocks,
       image_height_macroblocks);
    CUDA_ERRCK

    larger_sad_calc_16<<<macroblock_grid, dim3(32, 1)>>>
      (d_sads,
       image_width_macroblocks,
       image_height_macroblocks);
    CUDA_ERRCK

    //pb_SwitchToTimer(&timers, pb_TimerID_COPY);

    /* Transfer SAD data to the host */
    cudaMemcpy(sads_computed,// + 25 * MAX_POS_PADDED * image_size_macroblocks,
	       d_sads,// + 25 * MAX_POS_PADDED * image_size_macroblocks,
	       41 * MAX_POS_PADDED * image_size_macroblocks * sizeof(unsigned short)
,
           cudaMemcpyDeviceToHost);
    CUDA_ERRCK

    /* Free GPU memory */
    cudaFree(d_sads);
    CUDA_ERRCK
    cudaUnbindTexture(get_ref());
    CUDA_ERRCK
    cudaFreeArray(ref_ary);
    CUDA_ERRCK
    cudaFree(d_cur_image);
    CUDA_ERRCK

    //pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
  }

  /* Print output */
  if (params->outFile)
    {
      //pb_SwitchToTimer(&timers, pb_TimerID_IO);
      write_sads(params->outFile,
		 image_width_macroblocks,
		 image_height_macroblocks,
		 sads_computed);
      //pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
    }

#if 0  /* Debugging */
  print_test_sads(sads_computed, image_size_macroblocks);
  write_sads_directly("sad-debug.bin",
		      ref_image->width / 16, ref_image->height / 16,
		      sads_computed);
#endif

  /* Free memory */
  free(sads_computed);
  free_image(ref_image);
  free_image(cur_image);

  //pb_SwitchToTimer(&timers, pb_TimerID_NONE);
  //pb_PrintTimerSet(&timers);
  //pb_FreeParameters(params);

  return 0;
}


unsigned short
read16u(FILE *f)
{
  int n;

  n = fgetc(f);
  n += fgetc(f) << 8;

  return n;
}

short
read16i(FILE *f)
{
  int n;

  n = fgetc(f);
  n += fgetc(f) << 8;

  return n;
}

void
write32u(FILE *f, unsigned int i)
{
  putc(i, f);
  putc(i >> 8, f);
  putc(i >> 16, f);
  putc(i >> 24, f);
}

void
write16u(FILE *f, unsigned short h)
{
  putc(h, f);
  putc(h >> 8, f);
}

void
write16i(FILE *f, short h)
{
  putc(h, f);
  putc(h >> 8, f);
}



struct image_i16 *
load_image(char *filename)
{
  FILE *infile;
  short *data;
  int w;
  int h;

  infile = fopen(filename, "r");

  if (!infile)
    {
      fprintf(stderr, "Cannot find file '%s'\n", filename);
      exit(-1);
    }

  /* Read image dimensions */
  w = read16u(infile);
  h = read16u(infile);

  /* Read image contents */
  data = (short *)malloc(w * h * sizeof(short));
  size_t temp = fread(data, sizeof(short), w * h, infile);
  temp += w;

  fclose(infile);

  /* Create the return data structure */
  {
    struct image_i16 *ret =
      (struct image_i16 *)malloc(sizeof(struct image_i16));
    ret->width = w;
    ret->height = h;
    ret->data = data;
    return ret;
  }
}

void
free_image(struct image_i16 *img)
{
  free(img->data);
  free(img);
}
